from __future__ import annotations

import pkg_resources
import pytest

from negmas.genius.ginfo import (
    ALL_PASSING_NEGOTIATORS_NO_UNCERTAINTY,
    TESTED_NEGOTIATORS,
)
from negmas.genius.gnegotiators import *
from negmas.genius.negotiator import GeniusNegotiator
from negmas.inout import Scenario
from negmas.sao.negotiators import ToughNegotiator
from tests.switches import NEGMAS_FASTRUN, NEGMAS_RUN_GENIUS

TIMELIMIT = 10
STEPLIMIT = 50

# ALL_NEGOTIATORS = ALL_PASSING_NEGOTIATORS_NO_UNCERTAINTY
ALL_NEGOTIATORS = TESTED_NEGOTIATORS

AGENTS_WITH_NO_AGREEMENT_ON_SAME_preferences = (
    # "agents.anac.y2015.Mercury.Mercury",
    # "parties.in4010.q12015.group19.Group19",
    # "parties.in4010.q12015.group6.Group6",
    # "agents.anac.y2015.xianfa.XianFaAgent",
    "agents.anac.y2016.agentsmith.AgentSmith2016",
    "agents.anac.y2014.AgentTD.AgentTD",
    "agents.anac.y2019.agentgg.AgentGG",
    "agents.anac.y2019.saga.SAGA",
    "agents.anac.y2019.sacra.SACRA",
    "agents.anac.y2016.agentlight.AgentLight",
    "agents.anac.y2017.agentkn.AgentKN",
    "agents.anac.y2011.AgentK2.Agent_K2",
)

DOMAINS = [
    "tests/data/ItexvsCypress",
    "tests/data/Laptop",
]

SKIP_CONDITION = NEGMAS_FASTRUN or not NEGMAS_RUN_GENIUS
# SKIP_CONDITION = False
STRICT_TEST = True


def do_run(
    agent_factory,
    base_folder,
    opponent_preferences,
    agent_preferences,
    agent_starts,
    opponent_factory,
    n_steps,
    time_limit,
):
    domain = Scenario.from_genius_folder(base_folder)
    if not domain:
        raise ValueError(f"Cannot open domain {base_folder}")
    neg = domain.make_session(n_steps=n_steps, time_limit=time_limit)
    if neg is None:
        raise ValueError(f"Failed to load domain from {base_folder}")
    opponent = opponent_factory(preferences=domain.ufuns[opponent_preferences])
    theagent = agent_factory(preferences=domain.ufuns[agent_preferences])
    if agent_starts:
        neg.add(theagent)
        neg.add(opponent)
    else:
        neg.add(opponent)
        neg.add(theagent)
    neg.run()
    return neg


def do_test_same_ufun(agent_factory, base_folder, n_steps, time_limit, n_trials=3):
    # check that it will get to an agreement sometimes if the same ufun
    # is used for both agents
    from random import randint

    for _ in range(n_trials):
        indx = randint(0, 1)
        neg = do_run(
            agent_factory,
            base_folder,
            indx,
            indx,
            False,
            agent_factory,
            n_steps,
            time_limit,
        )
        if neg.agreement is not None:
            break
    else:
        assert (
            False
        ), f"failed to get an agreement in {n_trials} trials even using the same ufun\n{neg.trace}"  # type: ignore It makes not sense to have n_trials == 0 so we are safe


def do_test_genius_agent(
    AgentFactory,
    must_agree_if_same_preferences=True,
    java_class_name=None,
    strict_test=STRICT_TEST,
):
    if java_class_name is not None:
        AgentFactory = lambda *args, **kwargs: GeniusNegotiator(
            *args, java_class_name=java_class_name, strict=strict_test, **kwargs
        )
        agent_class_name = java_class_name
    else:
        agent_class_name = AgentFactory.__name__
    # print(f"Running {AgentClass.__name__}")
    for domain in DOMAINS:
        base_folder = pkg_resources.resource_filename("negmas", resource_name=domain)

        # check that it can run without errors with two different ufuns
        for opponent_type in (ToughNegotiator, Atlas3):
            for starts in (False, True):
                for n_steps, time_limit in ((STEPLIMIT, None), (None, TIMELIMIT)):
                    for ufuns in ((1, 0), (0, 1)):
                        try:
                            do_run(
                                AgentFactory,
                                base_folder,
                                ufuns[0],
                                ufuns[1],
                                starts,
                                opponent_type,
                                n_steps=n_steps,
                                time_limit=time_limit,
                            )
                        except Exception as e:
                            print(
                                f"{agent_class_name} FAILED against {opponent_type.__name__}"
                                f" going {'first' if starts else 'last'} ({n_steps} steps with "
                                f"{time_limit} limit taking ufun {ufuns[1]}."
                            )
                            raise e

        if not must_agree_if_same_preferences or (
            java_class_name is None
            and AgentFactory in AGENTS_WITH_NO_AGREEMENT_ON_SAME_preferences
        ):
            continue
        do_test_same_ufun(AgentFactory, base_folder, STEPLIMIT, None, 3)

    # GeniusBridge.clean()


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
@pytest.mark.parametrize("negotiator", ALL_NEGOTIATORS)
def test_all_negotiators(negotiator):
    do_test_genius_agent(None, java_class_name=negotiator)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_boulware_party():
    do_test_genius_agent(
        None, java_class_name="negotiator.parties.BoulwareNegotiationParty"
    )


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_boulware_agent():
    do_test_genius_agent(None, java_class_name="agents.TimeDependentAgentBoulware")


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_FSEGA2019():
    do_test_genius_agent(FSEGA2019)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MINF():
    do_test_genius_agent(MINF)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SACRA():
    do_test_genius_agent(SACRA, must_agree_if_same_preferences=False)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SAGA():
    do_test_genius_agent(SAGA, must_agree_if_same_preferences=False)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SENGOKU():
    do_test_genius_agent(SENGOKU)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ABMPAgent2():
    do_test_genius_agent(ABMPAgent2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Agent33():
    do_test_genius_agent(Agent33)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Agent36():
    do_test_genius_agent(Agent36)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentBuyog():
    do_test_genius_agent(AgentBuyog)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentF():
    do_test_genius_agent(AgentF)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentFSEGA():
    do_test_genius_agent(AgentFSEGA)


# @pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
@pytest.mark.skip("Known failure")
def test_AgentGG():
    do_test_genius_agent(AgentGG)


# @pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
@pytest.mark.skip("Known failure")
def test_AgentGP():
    do_test_genius_agent(AgentGP)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentH():
    do_test_genius_agent(AgentH)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentHerb():
    do_test_genius_agent(AgentHerb)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentHP():
    do_test_genius_agent(AgentHP)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentHP2():
    do_test_genius_agent(AgentHP2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentI():
    do_test_genius_agent(AgentI)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentK():
    do_test_genius_agent(AgentK)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentK2():
    do_test_genius_agent(AgentK2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentKF():
    do_test_genius_agent(AgentKF)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentKN():
    do_test_genius_agent(AgentKN, must_agree_if_same_preferences=False)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentLarry():
    do_test_genius_agent(AgentLarry, must_agree_if_same_preferences=False)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentLG():
    do_test_genius_agent(AgentLG)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentLight():
    do_test_genius_agent(AgentLight, must_agree_if_same_preferences=False)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentM():
    do_test_genius_agent(AgentM)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentMR():
    do_test_genius_agent(AgentMR)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentNP1():
    do_test_genius_agent(AgentNP1)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentQuest():
    do_test_genius_agent(AgentQuest)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentSmith():
    do_test_genius_agent(AgentSmith)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentSmith2016():
    do_test_genius_agent(AgentSmith2016)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentTD():
    do_test_genius_agent(AgentTD)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentTRP():
    do_test_genius_agent(AgentTRP)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentW():
    do_test_genius_agent(AgentW)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentX():
    do_test_genius_agent(AgentX)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentYK():
    do_test_genius_agent(AgentYK)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgreeableAgent2018():
    do_test_genius_agent(AgreeableAgent2018)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AnacSampleAgent():
    do_test_genius_agent(AnacSampleAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AresParty():
    do_test_genius_agent(AresParty)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ArisawaYaki():
    do_test_genius_agent(ArisawaYaki)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Aster():
    do_test_genius_agent(Aster)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ATeamAgent():
    do_test_genius_agent(AteamAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Atlas():
    do_test_genius_agent(Atlas)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Atlas3():
    do_test_genius_agent(Atlas3)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Atlas32016():
    do_test_genius_agent(Atlas32016)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_BayesianAgent():
    do_test_genius_agent(BayesianAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_BetaOne():
    do_test_genius_agent(BetaOne)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Betaone():
    do_test_genius_agent(Betaone)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_BoulwareNegotiationParty():
    do_test_genius_agent(BoulwareNegotiationParty)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_BramAgent():
    do_test_genius_agent(BramAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_BramAgent2():
    do_test_genius_agent(BramAgent2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_BraveCat():
    do_test_genius_agent(BraveCat)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Caduceus():
    do_test_genius_agent(Caduceus)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_CaduceusDC16():
    do_test_genius_agent(CaduceusDC16)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ClockworkAgent():
    do_test_genius_agent(ClockworkAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ConcederNegotiationParty():
    do_test_genius_agent(ConcederNegotiationParty)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ConDAgent():
    do_test_genius_agent(ConDAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_CUHKAgent():
    do_test_genius_agent(CUHKAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_CUHKAgent2015():
    do_test_genius_agent(CUHKAgent2015)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_DandikAgent():
    do_test_genius_agent(DandikAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_DoNA():
    do_test_genius_agent(DoNA)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_DrageKnight():
    do_test_genius_agent(DrageKnight)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_E2Agent():
    do_test_genius_agent(E2Agent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_EAgent():
    do_test_genius_agent(EAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ExpRubick():
    do_test_genius_agent(ExpRubick)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Farma():
    do_test_genius_agent(Farma)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Farma17():
    do_test_genius_agent(Farma17)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Flinch():
    do_test_genius_agent(Flinch)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_FullAgent():
    do_test_genius_agent(FullAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_FunctionalAcceptor():
    do_test_genius_agent(FunctionalAcceptor)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_FuzzyAgent():
    do_test_genius_agent(FuzzyAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_GAgent():
    do_test_genius_agent(GAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Gahboninho():
    do_test_genius_agent(Gahboninho)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Gangester():
    do_test_genius_agent(Gangester)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Gangster():
    do_test_genius_agent(Gangster)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_GaravelAgent():
    do_test_genius_agent(GaravelAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Gin():
    do_test_genius_agent(Gin)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_GrandmaAgent():
    do_test_genius_agent(GrandmaAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Gravity():
    do_test_genius_agent(Gravity)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group1():
    do_test_genius_agent(Group1)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group1BOA():
    do_test_genius_agent(Group1BOA)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AI2014Group2():
    do_test_genius_agent(AI2014Group2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group3():
    do_test_genius_agent(Group3)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group3Q2015():
    do_test_genius_agent(Group3Q2015)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group4():
    do_test_genius_agent(Group4)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group5():
    do_test_genius_agent(Group5)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group6():
    do_test_genius_agent(Group6)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group7():
    do_test_genius_agent(Group7)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group8():
    do_test_genius_agent(Group8)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group9():
    do_test_genius_agent(Group9)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group10():
    do_test_genius_agent(Group10)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group11():
    do_test_genius_agent(Group11)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group12():
    do_test_genius_agent(Group12)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group13():
    do_test_genius_agent(Group13)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group14():
    do_test_genius_agent(Group14)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group15():
    do_test_genius_agent(Group15)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group16():
    do_test_genius_agent(Group16)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group17():
    do_test_genius_agent(Group17)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group18():
    do_test_genius_agent(Group18)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group19():
    do_test_genius_agent(Group19)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group20():
    do_test_genius_agent(Group20)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group21():
    do_test_genius_agent(Group21)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Group22():
    do_test_genius_agent(Group22)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_AgentNeo():
    do_test_genius_agent(AgentNeo)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_GroupY():
    do_test_genius_agent(GroupY)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_HardDealer():
    do_test_genius_agent(HardDealer)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_HardHeaded():
    do_test_genius_agent(HardHeaded)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_IAMcrazyHaggler():
    do_test_genius_agent(IAMcrazyHaggler)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_IAMhaggler():
    do_test_genius_agent(IAMhaggler)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_IAMhaggler2011():
    do_test_genius_agent(IAMhaggler2011)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_IAMhaggler2012():
    do_test_genius_agent(IAMhaggler2012)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Imitator():
    do_test_genius_agent(Imitator)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ImmediateAcceptor():
    do_test_genius_agent(ImmediateAcceptor)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_InoxAgent():
    do_test_genius_agent(InoxAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_IQSun2018():
    do_test_genius_agent(IQSun2018)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_JonnyBlack():
    do_test_genius_agent(JonnyBlack)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_KAgent():
    do_test_genius_agent(KAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_KakeSoba():
    do_test_genius_agent(KakeSoba)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_KGAgent():
    do_test_genius_agent(KGAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Lancelot():
    do_test_genius_agent(Lancelot)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Libra():
    do_test_genius_agent(Libra)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MadAgent():
    do_test_genius_agent(MadAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Mamenchis():
    do_test_genius_agent(Mamenchis)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MaxOops():
    do_test_genius_agent(MaxOops)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MeanBot():
    do_test_genius_agent(MeanBot)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MengWan():
    do_test_genius_agent(MengWan)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Mercury():
    do_test_genius_agent(Mercury)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MetaAgent():
    do_test_genius_agent(MetaAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MetaAgent2012():
    do_test_genius_agent(MetaAgent2012)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MetaAgent2013():
    do_test_genius_agent(MetaAgent2013)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Mosa():
    do_test_genius_agent(Mosa)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_MyAgent():
    do_test_genius_agent(MyAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Ngent():
    do_test_genius_agent(Ngent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_NiceTitForTat():
    do_test_genius_agent(NiceTitForTat)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Nozomi():
    do_test_genius_agent(Nozomi)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_OMACagent():
    do_test_genius_agent(OMACagent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_OptimalBidderSimple():
    do_test_genius_agent(OptimalBidderSimple)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ParsAgent():
    do_test_genius_agent(ParsAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ParsAgent2():
    do_test_genius_agent(ParsAgent2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ParsAgent3():
    do_test_genius_agent(ParsAgent3)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ParsCat():
    do_test_genius_agent(ParsCat)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ParsCat2():
    do_test_genius_agent(ParsCat2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_PhoenixParty():
    do_test_genius_agent(PhoenixParty)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_PNegotiator():
    do_test_genius_agent(PNegotiator)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_PokerFace():
    do_test_genius_agent(PokerFace)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_PonPokoAgent():
    do_test_genius_agent(PonPokoAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_PonPokoRampage():
    do_test_genius_agent(PonPokoRampage)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Q12015Group2():
    do_test_genius_agent(Q12015Group2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_RandomCounterOfferNegotiationParty():
    do_test_genius_agent(RandomCounterOfferNegotiationParty)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_RandomDance():
    do_test_genius_agent(RandomDance)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_RandomParty():
    do_test_genius_agent(RandomParty)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_RandomParty2():
    do_test_genius_agent(RandomParty2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Rubick():
    do_test_genius_agent(Rubick)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Seto():
    do_test_genius_agent(Seto)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ShahAgent():
    do_test_genius_agent(ShahAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Shiboy():
    do_test_genius_agent(Shiboy)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SimilarityAgent():
    do_test_genius_agent(SimilarityAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Simpatico():
    do_test_genius_agent(Simpatico)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SimpleAgent():
    do_test_genius_agent(SimpleAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SimpleAgent2017():
    do_test_genius_agent(SimpleAgent2017)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SlavaAgent():
    do_test_genius_agent(SlavaAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SMACAgent():
    do_test_genius_agent(SMACAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Sobut():
    do_test_genius_agent(Sobut)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SolverAgent():
    do_test_genius_agent(SolverAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Sontag():
    do_test_genius_agent(Sontag)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SouthamptonAgent():
    do_test_genius_agent(SouthamptonAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_SYAgent():
    do_test_genius_agent(SYAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TaxiBox():
    do_test_genius_agent(TaxiBox)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Terra():
    do_test_genius_agent(Terra)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TheFawkes():
    do_test_genius_agent(TheFawkes)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TheNegotiator():
    do_test_genius_agent(TheNegotiator)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TheNegotiatorReloaded():
    do_test_genius_agent(TheNegotiatorReloaded)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TheNewDeal():
    do_test_genius_agent(TheNewDeal)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TimeDependentAgentBoulware():
    do_test_genius_agent(TimeDependentAgentBoulware)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TimeDependentAgentConceder():
    do_test_genius_agent(TimeDependentAgentConceder)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TimeDependentAgentHardliner():
    do_test_genius_agent(TimeDependentAgentHardliner)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TimeDependentAgentLinear():
    do_test_genius_agent(TimeDependentAgentLinear)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TMFAgent():
    do_test_genius_agent(TMFAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TucAgent():
    do_test_genius_agent(TucAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TUDelftGroup2():
    do_test_genius_agent(TUDelftGroup2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_TUDMixedStrategyAgent():
    do_test_genius_agent(TUDMixedStrategyAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_UtilityBasedAcceptor():
    do_test_genius_agent(UtilityBasedAcceptor)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_ValueModelAgent():
    do_test_genius_agent(ValueModelAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_WhaleAgent():
    do_test_genius_agent(WhaleAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_WinkyAgent():
    do_test_genius_agent(WinkyAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_XianFaAgent():
    do_test_genius_agent(XianFaAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Y2015Group2():
    do_test_genius_agent(Y2015Group2)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Yeela():
    do_test_genius_agent(Yeela)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Yushu():
    do_test_genius_agent(Yushu)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_YXAgent():
    do_test_genius_agent(YXAgent)


@pytest.mark.skipif(condition=SKIP_CONDITION, reason="Either no genius or fast run")
def test_Kawaii():
    do_test_genius_agent(Kawaii)


if __name__ == "__main__":
    pytest.main(args=[__file__])



from __future__ import annotations

import sys
from pathlib import Path
from time import perf_counter

import matplotlib
import typer
from matplotlib import functools
from matplotlib import pyplot as plt
from pandas.core.window.numba_ import Callable
from rich import print
from stringcase import titlecase

from negmas.genius.ginfo import get_java_class
from negmas.genius.negotiator import GeniusNegotiator
from negmas.helpers import get_class
from negmas.helpers.strings import camel_case, humanize_time, shortest_unique_names
from negmas.helpers.types import get_full_type_name
from negmas.inout import Scenario
from negmas.mechanisms import Mechanism
from negmas.negotiators.negotiator import Negotiator
from negmas.preferences.ops import (
    calc_reserved_value,
    kalai_points,
    make_rank_ufun,
    max_welfare_points,
    nash_points,
    pareto_frontier,
)

app = typer.Typer()

GENIUSMARKER = "genius"


def get_screen_resolution() -> tuple[int, int]:
    from tkinter import Tk

    # creating tkinter window
    root = Tk()
    # getting screen's height in pixels
    height = root.winfo_screenheight()
    # getting screen's width in pixels
    width = root.winfo_screenwidth()
    return (width, height)


def get_protocol(name: str) -> type[Mechanism]:
    if name.lower() == "sao":
        return get_class("negmas.sao.mechanism.SAOMechanism")
    if name.lower() == "tau":
        return get_class("negmas.gb.mechanisms.TAUMechanism")
    if name.lower() == "gtau":
        return get_class("negmas.gb.mechanisms.GeneralizedTAUMechanism")
    if name.lower() == "gao":
        return get_class("negmas.gb.mechanisms.GAOMechanism")
    if "." not in name:
        name = f"negmas.{name}"
    return get_class(name)


def get_genius_proper_class_name(s: str) -> str:
    assert s.startswith(GENIUSMARKER)
    return s.split(GENIUSMARKER)[-1]


def create_adapter(adapter_type, negotiator_type, name):
    return adapter_type(name=name, base=negotiator_type(name=name))


def make_genius_negotiator(*args, java_class_name: str, **kwargs):
    return GeniusNegotiator(*args, **kwargs, java_class_name=java_class_name)


def get_negotiator(class_name: str) -> type[Negotiator] | Callable[[str], Negotiator]:
    if class_name.startswith(GENIUSMARKER):
        for sp in (".", ":"):
            x = sp.join(class_name.split(sp)[1:])
            if x:
                class_name = x
                break
        java_class = get_java_class(class_name)
        if java_class is None:
            raise ValueError(
                f"Cannot find java class name for genius negotiator of type {class_name}"
            )
        return functools.partial(make_genius_negotiator, java_class_name=java_class)
    if "/" in class_name:
        adapter_name, _, negotiator_name = class_name.partition("/")
        adapter_type = get_adapter(adapter_name)
        negotiator_type = get_negotiator(negotiator_name)
        return functools.partial(create_adapter, adapter_type, negotiator_type)
    if "." not in class_name:
        if "_" in class_name:
            class_name = titlecase(camel_case(class_name))
        try:
            return get_class(f"negmas.{class_name}")
        except:
            if not class_name.endswith("Negotiator"):
                class_name = f"{class_name}Negotiator"
            class_name = f"negmas.{class_name}"
    return get_class(class_name)


def get_adapter(
    name: str, base_name="NegotiatorAdapter"
) -> type[Negotiator] | Callable[[str], Negotiator]:
    if "." not in name:
        if "_" in name:
            name = titlecase(camel_case(name))
        if not name.endswith(base_name):
            name = f"{name}{base_name}"
        name = f"negmas.gb.adapters.tau.{name}"
    return get_class(name)


def shorten_protocol_name(name: str) -> str:
    name = name.split(".")[-1]
    return name.replace("Mechanism", "").replace("Protocol", "")


def dist(x: tuple[float, ...], lst: list[tuple[float, ...]]):
    if not lst:
        return float("nan")
    return min(sum((a - b) ** 2 for a, b in zip(x, p, strict=True)) for p in lst)


def diff(x: tuple[float, ...], lst: list[tuple[float, ...]]):
    if not lst:
        return float("nan")
    s = sum(x)
    return min(abs(sum(_) - s) for _ in lst)


@app.command()
def run(
    domain: Path,
    protocol: str = typer.Option(
        "SAO",
        "--protocol",
        "--mechanism",
        "-p",
        "-m",
        help="The protocol (Mechanism to use)",
    ),
    negotiators: list[str] = typer.Option(
        ["aspiration", "aspiration"],
        "--agent",
        "--negotiator",
        "-n",
        "-a",
        help="Negotiator (agent) type. To use an adapter type, put the adapter name first separated from the negotiator name by a slash (e.g. TAUAdapter/AspirationNegotiator)",
    ),
    path: list[Path] = list(),
    reserved: list[float] = typer.Option(None, "--reserved", "-r"),
    fraction: list[float] = typer.Option(None, "--fraction", "-f"),
    normalize: bool = True,
    steps: int = typer.Option(None, "--steps", "-s"),  # type: ignore
    timelimit: int = typer.Option(None, "--time", "--timelimit", "-t"),  # type: ignore
    plot: bool = True,
    fast: bool = None,  # type: ignore
    plot_path: Path = None,  # type: ignore
    verbose: bool = typer.Option(False, "--verbose", "-v"),
    verbosity: int = 0,
    progress: bool = True,
    plot_backend: str = "",
    plot_interactive: bool = True,
    history: bool = False,
    stats: bool = True,
    rank_stats: bool = None,  # type: ignore
    discount: bool = True,
    extend_negotiators: bool = False,
    truncate_ufuns: bool = False,
    only2d: bool = False,
    show_agreement: bool = False,
    show_pareto_distance: bool = True,
    show_nash_distance: bool = True,
    show_kalai_distance: bool = True,
    show_max_welfare_distance: bool = True,
    show_max_relative_welfare_distance: bool = False,
    show_end_reason: bool = True,
    show_annotations: bool = False,
    show_reserved: bool = True,
    show_total_time: bool = True,
    show_relative_time: bool = True,
    show_n_steps: bool = True,
    simple_offers_view: bool = None,  # type: ignore
    raise_exceptions: bool = False,
    extra_params: str = "",
):
    kwargs = dict()
    if extra_params:
        extra_params = "dict(" + extra_params + ")"
        kwargs = eval(extra_params)

    if verbose and verbosity < 1:
        verbosity = 1
    for p in path:
        sys.path.append(str(p))
    adapter_names = shortest_unique_names(
        [_.split("/")[0] if "/" in _ else "" for _ in negotiators]
    )
    steps: int | float
    timelimit: int | float
    if steps is None:
        steps = float("inf")
    if timelimit is None:
        timelimit = float("inf")
    scenario = Scenario.from_genius_folder(
        domain, ignore_reserved=False, ignore_discount=not discount
    )
    if not scenario:
        print(f"Failed to load scenario from {domain}")
        return
    if normalize:
        scenario.normalize()

    assert scenario is not None
    scenario.mechanism_type = get_protocol(protocol)
    if verbosity > 0:
        print(f"Scenario: {domain}")
        print(
            f"Mechanism: {shorten_protocol_name(get_full_type_name(scenario.mechanism_type))}"
        )
        print(f"steps: {steps}\ntimelimit: {timelimit}")
    if (
        truncate_ufuns
        and len(scenario.ufuns) > len(negotiators)
        and len(negotiators) > 1
    ):
        scenario.ufuns = scenario.ufuns[: len(negotiators)]

    if (
        extend_negotiators
        and len(negotiators) > 0
        and len(negotiators) != len(scenario.ufuns)
    ):
        if len(negotiators) < len(scenario.ufuns):
            if verbosity > 0:
                print(
                    f"Found {len(scenario.ufuns)} ufuns and {len(negotiators)} negotiators. Will add negotiators of the last type to match the n. ufuns"
                )
            negotiators = negotiators + (
                [negotiators[-1]] * (len(scenario.ufuns) - len(negotiators))
            )

        if len(negotiators) > len(scenario.ufuns):
            if verbosity > 0:
                print(
                    f"Found {len(scenario.ufuns)} ufuns and {len(negotiators)} negotiators. Will ignore the last n. negotiators"
                )
            negotiators = negotiators[: len(scenario.ufuns)]

    negotiator_names = shortest_unique_names(negotiators, guarantee_unique=True)
    agents = [
        get_negotiator(_)(name=name) for _, name in zip(negotiators, negotiator_names, strict=True)  # type: ignore
    ]
    if len(agents) < 2:
        print(
            f"At least 2 negotiators are needed: found {[_.__class__.__name__ for _ in agents]}"
        )
        return
    if reserved and not extend_negotiators:
        assert len(reserved) == len(negotiators), f"{reserved=} but {negotiators=}"
    if reserved:
        for u, r in zip(scenario.ufuns, reserved):
            u.reserved_value = r
    if fraction:
        if len(fraction) < len(negotiators):
            fraction += [1.0] * (len(negotiators) - len(fraction))
        for u, f in zip(scenario.ufuns, fraction):
            u.reserved_value = calc_reserved_value(u, f)
    if (
        not extend_negotiators
        and len(agents) > 0
        and len(agents) != len(scenario.ufuns)
    ):
        print(
            f"You passed {len(agents)} agents for a negotiation with {len(scenario.ufuns)} ufuns. pass --extend-negotiators to adjust the agent number"
        )
        exit(1)
    session = scenario.make_session(
        agents,
        n_steps=steps,
        time_limit=timelimit,
        verbosity=verbosity - 1,
        **dict(ignore_negotiator_exceptions=not raise_exceptions),
        **kwargs,
    )
    if len(session.negotiators) < 2:
        print(
            f"At least 2 negotiators are needed: Only the following could join {[_.__class__.__name__ for _ in session.negotiators]}"
        )
        return
    if verbosity > 0:
        print(f"Adapters: {', '.join(adapter_names)}")
    if verbosity > 1:
        print(f"Negotiators: {', '.join(negotiator_names)}")
    runner = session.run_with_progress if progress else session.run
    _start = perf_counter()
    state = runner()
    duration = perf_counter() - _start
    if verbosity > 1:
        print(f"Time: {humanize_time(duration, show_ms=True, show_us=True)}")
        print(f"Steps: {session.current_step}")
        print(state)
    print(f"Agreement: {state.agreement}")
    print(f"Utilities: {[u(state.agreement) for u in scenario.ufuns]}")
    print(
        f"Advantages: {[u(state.agreement) - (u.reserved_value if u.reserved_value is not None else float('inf')) for u in scenario.ufuns]}"
    )
    fast = fast or fast is None and scenario.outcome_space.cardinality > 10_000
    if fast:
        # rank_stats = stats = False
        if simple_offers_view is None:
            simple_offers_view = True
    rank_stats = (
        rank_stats or rank_stats is None and scenario.outcome_space.cardinality < 10_000
    )

    if stats or rank_stats:
        pareto, pareto_outcomes = session.pareto_frontier()
    else:
        pareto, pareto_outcomes = [], []
    if stats:
        utils = tuple(u(state.agreement) for u in scenario.ufuns)
        print(f"Agreement Utilities: {utils}")
        pts = session.nash_points(frontier=pareto, frontier_outcomes=pareto_outcomes)
        print(
            f"Nash Points: {pts} -- Distance = {dist(utils, list(a for a, _ in pts))}"
        )
        pts = session.kalai_points(frontier=pareto, frontier_outcomes=pareto_outcomes)
        print(
            f"Kalai Points: {pts} -- Distance = {dist(utils, list(a for a, _ in pts))}"
        )
        pts = session.modified_kalai_points(
            frontier=pareto, frontier_outcomes=pareto_outcomes
        )
        print(
            f"Modified Kalai Points: {pts} -- Distance = {dist(utils, list(a for a, _ in pts))}"
        )
        pts = session.max_welfare_points(
            frontier=pareto, frontier_outcomes=pareto_outcomes
        )
        print(
            f"Max. Welfare Points: {pts} -- Diff = {diff(utils, list(a for a, _ in pts))}"
        )
        # pts = session.max_relative_welfare_points(frontier=pareto, frontier_outcomes=pareto_outcomes)
        # print(
        #     f"Max. Relative Points: {pts} -- Distance = {dist(utils, list(a for a, _ in pts))}"
        # )
    if rank_stats:
        ranks_ufuns = tuple(make_rank_ufun(_) for _ in scenario.ufuns)
        ranks = tuple(_(state.agreement) for _ in ranks_ufuns)
        print(f"Agreement Relative Ranks: {ranks}")
        pareto_save = pareto_outcomes
        alloutcomes = session.discrete_outcomes()
        pareto, pareto_indices = pareto_frontier(
            ranks_ufuns, outcomes=alloutcomes, sort_by_welfare=True
        )
        pareto_outcomes = [alloutcomes[_] for _ in pareto_indices]
        if len(pareto_save) != len(pareto_outcomes) or any(
            a != b
            for a, b in zip(sorted(pareto_save), sorted(pareto_outcomes), strict=True)
        ):
            print(
                f"[bold red]Ordinal pareto outcomes do not match cardinal pareto outcomes[/bold red]\nOrdinal: {pareto_outcomes}\nCardinal: {pareto_save}"
            )
        utils_indices = nash_points(
            frontier=pareto, ufuns=ranks_ufuns, outcome_space=scenario.outcome_space
        )
        pts = tuple((a, pareto_outcomes[b]) for a, b in utils_indices)
        nutils = list(a for a, _ in utils_indices)
        print(f"Ordinal Nash Points: {pts} -- Distance = {dist(ranks, nutils)}")
        pts = kalai_points(
            frontier=pareto,
            ufuns=ranks_ufuns,
            outcome_space=scenario.outcome_space,
            subtract_reserved_value=True,
        )
        pts = tuple((a, pareto_outcomes[b]) for a, b in utils_indices)
        nutils = list(a for a, _ in utils_indices)
        print(f"Ordinal Kalai Points: {pts} -- Distance = {dist(ranks, nutils)}")
        pts = kalai_points(
            frontier=pareto,
            ufuns=ranks_ufuns,
            outcome_space=scenario.outcome_space,
            subtract_reserved_value=False,
        )
        pts = tuple((a, pareto_outcomes[b]) for a, b in utils_indices)
        nutils = list(a for a, _ in utils_indices)
        print(
            f"Ordinal Modified Kalai Points: {pts} -- Distance = {dist(ranks, nutils)}"
        )
        pts = max_welfare_points(
            frontier=pareto, ufuns=ranks_ufuns, outcome_space=scenario.outcome_space
        )
        pts = tuple((a, pareto_outcomes[b]) for a, b in utils_indices)
        nutils = list(a for a, _ in utils_indices)
        print(f"Ordinal Max. Welfare Points: {pts} -- Diff = {diff(ranks, nutils)}")
        # pts = max_relative_welfare_points(
        #     frontier=pareto, ufuns=ranks_ufuns, outcome_space=scenario.outcome_space
        # )
        # pts = tuple((a, pareto_outcomes[b]) for a, b in utils_indices)
        # nutils = list(a for a, _ in utils_indices)
        # print(f"Ordinal Max. Relative Points: {pts} -- Distance = {dist(ranks, nutils)}")
    if history:
        if hasattr(session, "full_trace"):
            print(session.full_trace)  # type: ignore full_trace is defined for SAO and GBM
        else:
            print(session.history)
    if plot_path:
        plot_path = Path(plot_path).absolute()
        plot_path.parent.mkdir(parents=True, exist_ok=True)
    if plot:
        if plot_backend:
            matplotlib.use(plot_backend)
            matplotlib.interactive(plot_interactive)
        session.plot(
            save_fig=plot_path is not None,
            path=plot_path.parent if plot_path else None,
            fig_name=plot_path.name if plot_path else None,
            only2d=only2d,
            show_agreement=show_agreement,
            show_pareto_distance=show_pareto_distance,
            show_nash_distance=show_nash_distance,
            show_kalai_distance=show_kalai_distance,
            show_max_welfare_distance=show_max_welfare_distance,
            show_max_relative_welfare_distance=show_max_relative_welfare_distance,
            show_end_reason=show_end_reason,
            show_annotations=show_annotations,
            show_reserved=show_reserved,
            show_total_time=show_total_time,
            show_relative_time=show_relative_time,
            show_n_steps=show_n_steps,
            fast=fast,
            simple_offers_view=simple_offers_view,
        )
        mng = plt.get_current_fig_manager()
        mng.resize(1024, 860)
        mng.full_screen_toggle()
        plt.show()


if __name__ == "__main__":
    app()



from __future__ import annotations

from negmas.inout import Scenario
from negmas.outcomes import make_issue
from negmas.outcomes.outcome_space import make_os
from negmas.preferences import LinearAdditiveUtilityFunction as U
from negmas.sao import AspirationNegotiator, RandomNegotiator
from negmas.sao.mechanism import SAOMechanism
from negmas.tournaments.neg import cartesian_tournament


def test_can_run_cartesian_simple_tournament_n_reps():
    n = 2
    rotate_ufuns = True
    n_repetitions = 4
    issues = (
        make_issue([f"q{i}" for i in range(10)], "quantity"),
        make_issue([f"p{i}" for i in range(5)], "price"),
    )
    ufuns = [
        (
            U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
            U.random(issues=issues, reserved_value=(0.0, 0.2), normalized=False),
        )
        for _ in range(n)
    ]
    scenarios = [
        Scenario(
            outcome_space=make_os(issues, name=f"S{i}"),
            ufuns=u,
            mechanism_type=SAOMechanism,  # type: ignore
            mechanism_params=dict(),
        )
        for i, u in enumerate(ufuns)
    ]
    competitors = [RandomNegotiator, AspirationNegotiator]
    results = cartesian_tournament(
        competitors=competitors,
        scenarios=scenarios,
        mechanism_params=dict(n_steps=10),
        n_repetitions=n_repetitions,
        verbosity=0,
        rotate_ufuns=rotate_ufuns,
        # plot_fraction=0.5,
        path=None,
    )
    scores = results.scores
    assert (
        len(scores)
        == len(scenarios)
        * (int(rotate_ufuns) + 1)  # two variations if rotate_ufuns else one
        * len(competitors)
        * len(competitors)
        * n_repetitions
        * 2  # two scores per run
    )

    assert (
        len(results.details)
        == len(scenarios)
        * (int(rotate_ufuns) + 1)  # two variations if rotate_ufuns else one
        * len(competitors)
        * len(competitors)
        * n_repetitions
    )
    assert len(results.final_scores) == len(competitors)



from __future__ import annotations

import random

import matplotlib.pyplot as plt
import numpy as np
import pytest

from negmas import (
    AspirationNegotiator,
    FirstOfferOrientedTBNegotiator,
    NaiveTitForTatNegotiator,
    SAOController,
    SAOMechanism,
    TopFractionNegotiator,
    ToughNegotiator,
)
from negmas.outcomes.base_issue import make_issue
from negmas.preferences import MappingUtilityFunction, RandomUtilityFunction
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.preferences.value_fun import AffineFun, IdentityFun, LinearFun

random.seed(0)
np.random.seed(0)


def test_tough_asp_negotiator():
    a1 = ToughNegotiator()
    a2 = AspirationNegotiator(aspiration_type="conceder")
    outcomes = [(_,) for _ in range(10)]
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    neg = SAOMechanism(outcomes=outcomes, n_steps=100)
    neg.add(a1, preferences=u1)
    neg.add(a2, preferences=u2)
    neg.run()
    a1offers = neg.negotiator_offers(a1.id)
    a2offers = neg.negotiator_offers(a2.id)
    if len(a1offers) > 0:
        assert len(set(a1offers)) == 1 and a1offers[-1] == (9,)
    assert len(set(a2offers)) >= 0


def test_tough_tit_for_tat_negotiator():
    a1 = ToughNegotiator()
    a2 = NaiveTitForTatNegotiator()
    outcomes = [(_,) for _ in range(10)]
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    neg = SAOMechanism(outcomes=outcomes, n_steps=100)
    neg.add(a1, preferences=u1)
    neg.add(a2, preferences=u2)
    neg.run()
    a1offers = neg.negotiator_offers(a1.id)
    a2offers = neg.negotiator_offers(a2.id)
    # print(a1offers)
    # print(a2offers)
    if len(a1offers) > 0:
        assert len(set(a1offers)) == 1 and a1offers[-1] == (9,)
    assert len(set(a2offers)) >= 0


def test_asp_negotaitor():
    a1 = AspirationNegotiator(name="a1")
    a2 = AspirationNegotiator(name="a2")
    outcomes = [(_,) for _ in range(100)]
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    neg = SAOMechanism(outcomes=outcomes, n_steps=100)
    neg.add(a1, preferences=u1)
    neg.add(a2, preferences=u2)
    neg.run()
    a1offers = neg.negotiator_offers(a1.id)
    a2offers = neg.negotiator_offers(a2.id)
    assert a1offers[0] == (99,)
    assert a2offers[0] == (0,)
    for i, offer in enumerate(_[0] for _ in a1offers):
        assert i == 0 or offer <= a1offers[i - 1][0]
    for i, offer in enumerate(_[0] for _ in a2offers):
        assert i == 0 or offer >= a2offers[i - 1][0]
    assert neg.state.agreement is not None
    assert neg.state.agreement in ((49,), (50,))


@pytest.mark.skip(
    reason="TFT against itself will become hard-headed. A general solution is to add an offering strategy that breaks ties and combine it with them"
)
def test_tit_for_tat_negotiators_agree_in_the_middle():
    a1 = NaiveTitForTatNegotiator(name="a1")
    a2 = NaiveTitForTatNegotiator(name="a2", kindness=0.01)
    outcomes = [(_,) for _ in range(100)]
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    neg = SAOMechanism(outcomes=outcomes, n_steps=100)
    neg.add(a1, preferences=u1)
    neg.add(a2, preferences=u2)
    neg.run()
    a1offers = neg.negotiator_offers(a1.id)
    a2offers = neg.negotiator_offers(a2.id)
    # print(a1offers)
    # print(a2offers)
    assert a1offers[0] == (99,), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    assert a2offers[0] == (0,), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    assert (
        neg.state.agreement is not None
    ), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    assert (
        40 <= neg.state.agreement[0] <= 60
    ), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    # for i, offer in enumerate(_[0] for _ in a1offers):
    #     assert i == 0 or offer <= a1offers[i - 1][0] + 2, f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    # for i, offer in enumerate(_[0] for _ in a2offers):
    #     assert i == 0 or offer >= a2offers[i - 1][0] + 2, f"{neg.plot()}{plt.show()}{neg.extended_trace}"


def test_top_only_negotiator():
    outcomes = [(_,) for _ in range(10)]
    a1 = ToughNegotiator(name="a1")
    a2 = ToughNegotiator(name="a2")
    u1 = 22.0 - np.linspace(0.0, 22.0, len(outcomes))
    neg = SAOMechanism(outcomes=outcomes, n_steps=10)
    neg.add(
        a1,
        preferences=MappingUtilityFunction(
            dict(zip(outcomes, u1)), outcome_space=neg.outcome_space
        ),
    )
    neg.add(
        a2,
        preferences=MappingUtilityFunction(
            dict(zip(outcomes, 22 - u1)), outcome_space=neg.outcome_space
        ),
    )
    neg.run()
    assert neg.state.timedout
    first = neg.negotiator_offers(neg.negotiator_ids[0])
    second = neg.negotiator_offers(neg.negotiator_ids[1])
    assert len(set(first)) == 1
    assert len(set(second)) == 1
    assert first[0] == (0,)
    assert second[0] == (9,)


def test_tft_propose():
    outcomes = [(_,) for _ in range(10)]
    a1 = NaiveTitForTatNegotiator(name="a1", initial_concession="min")
    a2 = ToughNegotiator(name="a2")
    u1 = 22.0 - np.linspace(0.0, 22.0, len(outcomes))
    neg = SAOMechanism(outcomes=outcomes, n_steps=10)
    neg.add(
        a1,
        preferences=MappingUtilityFunction(
            dict(zip(outcomes, u1)), outcome_space=neg.outcome_space
        ),
    )
    neg.add(
        a2,
        preferences=MappingUtilityFunction(
            dict(zip(outcomes, 22 - u1)), outcome_space=neg.outcome_space
        ),
    )
    neg.step()
    proposal = neg.negotiator_offers(neg.negotiators[0].id)[0]
    assert proposal == (0,), "Proposes top first"

    neg.step()
    proposal = neg.negotiator_offers(neg.negotiators[0].id)[1]
    assert proposal == (1,), "Proposes second second if min concession is set"

    a1 = NaiveTitForTatNegotiator(name="a1")
    a2 = ToughNegotiator(name="a1")
    u1 = [50.0] * 3 + (22 - np.linspace(10.0, 22.0, len(outcomes) - 3)).tolist()
    neg = SAOMechanism(outcomes=outcomes, n_steps=10)
    neg.add(
        a1,
        preferences=MappingUtilityFunction(lambda x: u1[x[0]], outcomes=outcomes),
    )
    neg.add(
        a2,
        preferences=MappingUtilityFunction(lambda x: 22 - u1[x[0]], outcomes=outcomes),
    )

    neg.step()
    proposal = neg.negotiator_offers(neg.negotiators[0].id)[-1]
    assert proposal in ((0,), (1,), (2,)), "Proposes top first"

    neg.step()
    proposal = neg.negotiator_offers(neg.negotiators[0].id)[-1]
    assert proposal in (
        (0,),
        (1,),
        (2,),
        (3,),
    ), "Proposes first item with utility less than the top if concession is min"


def test_tit_for_tat_against_asp_negotiators():
    a1 = NaiveTitForTatNegotiator(name="tft")
    a2 = AspirationNegotiator(name="asp")
    outcomes = [(_,) for _ in range(10)]
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    neg = SAOMechanism(outcomes=outcomes, n_steps=20, time_limit=None)
    neg.add(a1, preferences=u1)
    neg.add(a2, preferences=u2)
    neg.run()
    a1offers = neg.negotiator_offers(a1.id)
    a2offers = neg.negotiator_offers(a2.id)
    assert a1offers[0] == (9,), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    # assert a2offers[0] == (0,)
    for i, offer in enumerate(_[0] for _ in a2offers):
        assert (
            i == 0 or offer >= a2offers[i - 1][0]
        ), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    assert (
        neg.state.agreement is not None
    ), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    assert neg.state.agreement in (
        (1,),
        (2,),
        (3,),
        (4,),
        (5,),
        (6,),
    ), f"{neg.plot()}{plt.show()}{neg.extended_trace}"
    for i, offer in enumerate(_[0] for _ in a1offers):
        assert (
            i == 0 or offer <= a1offers[i - 1][0] + 2
        ), f"{neg.plot()}{plt.show()}{neg.extended_trace}"


def test_best_only_asp_negotiator():
    a1 = TopFractionNegotiator(min_utility=0.9, top_fraction=0.09)
    a2 = AspirationNegotiator(aspiration_type="conceder")
    outcomes = [(_,) for _ in range(20)]
    u1 = MappingUtilityFunction(
        dict(zip(outcomes, np.linspace(0.0, 1.0, len(outcomes)).tolist())),
        outcomes=outcomes,
    )
    u2 = MappingUtilityFunction(
        dict(zip(outcomes, (1 - np.linspace(0.0, 1.0, len(outcomes))).tolist())),
        outcomes=outcomes,
    )
    neg = SAOMechanism(outcomes=outcomes, n_steps=200)
    neg.add(a1, preferences=u1)
    neg.add(a2, preferences=u2)
    neg.run()
    a1offers = neg.negotiator_offers(a1.id)
    a2offers = neg.negotiator_offers(a2.id)
    if len(a1offers) > 0:
        assert (
            len(set(a1offers)) <= 2
            and min(u1(_) for _ in a1offers if _ is not None) >= 0.9
        )
    assert len(set(a2offers)) >= 1


def test_controller():
    n_sessions = 5
    c = SAOController(
        default_negotiator_type="negmas.sao.AspirationNegotiator",
        default_negotiator_params={"aspiration_type": "conceder"},
    )
    sessions = [SAOMechanism(outcomes=10, n_steps=10) for _ in range(n_sessions)]
    for session in sessions:
        session.add(
            AspirationNegotiator(aspiration_type="conceder"),
            preferences=RandomUtilityFunction(outcome_space=session.outcome_space),
        )
        session.add(
            c.create_negotiator(),
            preferences=RandomUtilityFunction(outcome_space=session.outcome_space),
        )
    completed: list[int] = []
    while len(completed) < n_sessions:
        for i, session in enumerate(sessions):
            if i in completed:
                continue
            state = session.step()
            if state.broken or state.timedout or state.agreement is not None:
                completed.append(i)
    # we are just checking that the controller runs. No need to assert anything


def test_negotiator_checkpoint():
    pass


def test_smart_aspiration():
    # create negotiation agenda (issues)
    issues = [
        make_issue(name="price", values=10),
        make_issue(name="quantity", values=(1, 11)),
        make_issue(name="delivery_time", values=10),
    ]

    # create the mechanism
    session = SAOMechanism(issues=issues, n_steps=20)

    # define ufuns
    seller_utility = LinearAdditiveUtilityFunction(
        values={
            "price": IdentityFun(),
            "quantity": LinearFun(0.2),
            "delivery_time": AffineFun(-1, bias=9),
        },
        weights={"price": 1.0, "quantity": 1.0, "delivery_time": 10.0},
        outcome_space=session.outcome_space,
    ).scale_max(1.0)
    buyer_utility = LinearAdditiveUtilityFunction(
        values={
            "price": AffineFun(-1, bias=9.0),
            "quantity": LinearFun(0.2),
            "delivery_time": IdentityFun(),
        },
        outcome_space=session.outcome_space,
    ).scale_max(1.0)

    session.add(FirstOfferOrientedTBNegotiator(name="buyer"), ufun=buyer_utility)
    session.add(FirstOfferOrientedTBNegotiator(name="seller"), ufun=seller_utility)
    session.run()


def test_outcome_space_setting_resetting():
    # create negotiation agenda (issues)
    issues = [
        make_issue(name="price", values=10),
        make_issue(name="quantity", values=(1, 11)),
        make_issue(name="delivery_time", values=10),
    ]

    # create the mechanism
    session = SAOMechanism(issues=issues, n_steps=20)

    # define ufuns
    seller_utility = LinearAdditiveUtilityFunction(
        values=(
            IdentityFun(),
            LinearFun(0.2),
            AffineFun(-1, bias=9),
        ),
        weights=(1.0, 1.0, 10.0),
    )
    buyer_utility = LinearAdditiveUtilityFunction(
        values=(
            AffineFun(-1, bias=9.0),
            LinearFun(0.2),
            IdentityFun(),
        ),
    )
    assert seller_utility.outcome_space is None
    assert buyer_utility.outcome_space is None

    session.add(FirstOfferOrientedTBNegotiator(name="buyer"), ufun=buyer_utility)
    assert seller_utility.outcome_space is None
    assert buyer_utility.outcome_space is None
    session.add(FirstOfferOrientedTBNegotiator(name="seller"), ufun=seller_utility)
    assert seller_utility.outcome_space is None
    assert buyer_utility.outcome_space is None
    session.step()
    assert seller_utility.outcome_space is not None
    assert buyer_utility.outcome_space is not None
    session.run()
    assert seller_utility.outcome_space is None
    assert buyer_utility.outcome_space is None


if __name__ == "__main__":
    pytest.main(args=[__file__])



from __future__ import annotations

from random import shuffle
from time import perf_counter
from typing import TYPE_CHECKING, Any, Callable

from attrs import define

from negmas.common import TraceElement
from negmas.helpers import humanize_time
from negmas.helpers.types import get_full_type_name, instantiate
from negmas.mechanisms import Mechanism, MechanismStepResult
from negmas.outcomes import Outcome
from negmas.plots.util import default_colorizer
from negmas.preferences import BaseUtilityFunction, Preferences

from ..common import GBAction, GBResponse, GBState, ResponseType, ThreadState
from ..constraints.base import LocalOfferingConstraint, OfferingConstraint
from ..evaluators.base import EvaluationStrategy, LocalEvaluationStrategy, all_accept

if TYPE_CHECKING:
    from negmas.gb.negotiators.base import GBNegotiator
    from negmas.plots.util import Colorizer

__all__ = ["GBMechanism", "ParallelGBMechanism", "SerialGBMechanism"]

DEFAULT_COLORMAP = "jet"


@define
class GBThread:
    mechanism: BaseGBMechanism
    negotiator: GBNegotiator
    responders: list[GBNegotiator]
    indx: int
    state: ThreadState
    evaluator: LocalEvaluationStrategy | None = None
    constraint: LocalOfferingConstraint | None = None

    @property
    def accepted_offers(self) -> list[Outcome]:
        return self.state.accepted_offers

    def run(
        self, action: dict[str, Outcome | None] | None = None
    ) -> tuple[ThreadState, GBResponse | None]:
        mechanism_state: GBState = self.mechanism.state  # type: ignore
        history: list[GBState] = self.mechanism.history  # type: ignore
        source = self.negotiator.id
        if action is None:
            offer = self.negotiator.propose(mechanism_state)
        else:
            offer = action.get(source, None)
        # assert offer is None or isinstance(offer, Outcome)
        self.state.new_offer = offer
        if self.constraint and not self.constraint(
            mechanism_state.threads[source],
            [_.threads[source] for _ in history],
        ):
            self.state.new_offer = offer = None

        if offer is None:
            if self.mechanism._extra_callbacks:
                for n in self.responders:
                    n.on_partner_ended(source)
            self.state.new_responses = dict()
            return (
                self.state,
                self.evaluator.eval(
                    source,
                    self.state,
                    mechanism_state.thread_history(history, source),
                    mechanism_state.base_state,
                )
                if self.evaluator
                else None,
            )
        if self.mechanism._extra_callbacks:
            for n in self.responders:
                n.on_partner_proposal(mechanism_state, source, offer)
        responses = [_.respond(mechanism_state, source=source) for _ in self.responders]
        if self.mechanism._extra_callbacks:
            for n, r in zip(self.responders, responses):
                n.on_partner_response(mechanism_state, n.id, offer, r)
        if all(_ == ResponseType.ACCEPT_OFFER for _ in responses):
            self.state.accepted_offers.append(offer)
        self.state.new_responses = dict(
            zip(tuple(_.id for _ in self.responders), responses)
        )

        return (
            self.state,
            self.evaluator.eval(
                source,
                self.state,
                mechanism_state.thread_history(history, source),
                mechanism_state.base_state,
            )
            if self.evaluator
            else None,
        )


class BaseGBMechanism(Mechanism):
    def __init__(
        self,
        *args,
        dynamic_entry=False,
        extra_callbacks=True,
        check_offers=False,
        enforce_issue_types=False,
        cast_offers=False,
        end_on_no_response=True,
        ignore_negotiator_exceptions=False,
        parallel: bool = True,
        sync_calls: bool = False,
        initial_state: GBState | None = None,
        **kwargs,
    ):
        super().__init__(
            *args,
            dynamic_entry=dynamic_entry,
            extra_callbacks=extra_callbacks,
            initial_state=GBState() if not initial_state else initial_state,
            **kwargs,
        )
        self._current_state: GBState
        self._parallel = parallel
        self._threads: list[GBThread] = []
        self._ignore_negotiator_exceptions = ignore_negotiator_exceptions
        self._dynamic_entry = dynamic_entry
        self._extra_callbacks = extra_callbacks
        self._check_offers = check_offers
        self._enforce_issue_types = enforce_issue_types
        self._cast_offers = cast_offers
        self._end_on_no_response = enforce_issue_types
        self._ignore_negotiator_exceptions = enforce_issue_types
        self._parallel = parallel
        self._sync_calls = sync_calls
        self.params["check_offers"] = check_offers
        self.params["enforce_issue_types"] = enforce_issue_types
        self.params["ignore_negotiator_exceptions"] = ignore_negotiator_exceptions
        self.params["cast_offers"] = cast_offers
        self.params["end_on_no_response"] = end_on_no_response
        self.params["enable_callbacks"] = extra_callbacks
        self.params["parallel"] = parallel
        self.params["sync_calls"] = sync_calls

    @property
    def state(self) -> GBState:
        """Returns the current state.

        Override `extra_state` if you want to keep extra state
        """
        return self._current_state

    def set_sync_call(self, v: bool):
        self._sync_call = v

    def run_threads(
        self, action: dict[str, Outcome | None] | None = None
    ) -> dict[str, tuple[ThreadState, GBResponse | None]]:
        def _do_run(idd, thread: GBThread):
            if self.verbosity > 2:
                print(
                    f"{self.name}: Thread {thread.negotiator.name} starts after {humanize_time(perf_counter() - self._start_time, show_ms=True) if self._start_time else 0}",
                    flush=True,
                )
            r = thread.run(action)
            state: GBState = self.state  # type: ignore
            state.last_thread = idd
            return r

        if not self._parallel:
            threads = self._threads
            return {
                thread.negotiator.id: _do_run(thread.negotiator.id, thread)
                for thread in threads
            }
        # todo: make this really parallel
        indices = [_ for _ in range(len(self._threads))]
        shuffle(indices)
        threads = [self._threads[_] for _ in indices]
        results = dict()
        for t in threads:
            results[t.negotiator.id] = _do_run(t.negotiator.id, t)
        return results

    @property
    def full_trace(self) -> list[TraceElement]:
        def response(state: GBState):
            if state.timedout:
                return "timedout"
            if state.ended:
                return "ended"
            if state.has_error:
                return "error"
            if state.agreement:
                return "accepted"
            return "rejected"

        offers = []
        self._history: list[GBState]  # type: ignore
        for state in self._history:
            state: GBState
            offers += [
                TraceElement(
                    state.time,
                    state.relative_time,
                    state.step,
                    source,
                    t.new_offer,
                    {neg: int(resp) for neg, resp in t.new_responses.items()},
                    response(state),
                )
                for source, t in state.threads.items()
            ]

        return offers

    @property
    def extended_trace(self) -> list[tuple[int, str, Outcome]]:
        """Returns the negotiation history as a list of step/negotiator/offer
        tuples."""
        offers = []
        self._history: list[GBState]  # type: ignore
        for state in self._history:
            offers += [
                (state.step, source, t.new_offer) for source, t in state.threads.items()
            ]
        return offers

    @property
    def trace(self) -> list[tuple[str, Outcome]]:
        """Returns the negotiation history as a list of negotiator/offer
        tuples."""
        offers = []
        for state in self._history:
            offers += [(source, t.new_offer) for source, t in state.threads.items()]

        return offers

    def negotiator_offers(self, negotiator_id: str) -> list[Outcome]:
        """Returns the offers given by a negotiator (in order)"""
        return [o for n, o in self.trace if n == negotiator_id]

    def negotiator_full_trace(
        self, negotiator_id: str
    ) -> list[tuple[float, float, int, Outcome, str]]:
        """Returns the (time/relative-time/step/outcome/response) given by a
        negotiator (in order)"""
        return [
            (t, rt, s, o, a)
            for t, rt, s, n, o, _, a in self.full_trace
            if n == negotiator_id
        ]

    @property
    def offers(self) -> list[Outcome]:
        """Returns the negotiation history as a list of offers."""
        return [o for _, o in self.trace]

    @property
    def _step(self):
        """A private property used by the checkpoint system."""
        return self._current_state.step

    def plot(
        self,
        plotting_negotiators: tuple[int, int] | tuple[str, str] = (0, 1),
        save_fig: bool = False,
        path: str | None = None,
        fig_name: str | None = None,
        ignore_none_offers: bool = True,
        with_lines: bool = True,
        show_agreement: bool = False,
        show_pareto_distance: bool = True,
        show_nash_distance: bool = True,
        show_kalai_distance: bool = True,
        show_max_welfare_distance: bool = True,
        show_max_relative_welfare_distance: bool = False,
        show_end_reason: bool = True,
        show_annotations: bool = False,
        show_reserved: bool = True,
        show_total_time=True,
        show_relative_time=True,
        show_n_steps=True,
        colors: list | None = None,
        markers: list[str] | None = None,
        colormap: str = DEFAULT_COLORMAP,
        ylimits: tuple[float, float] | None = None,
        common_legend: bool = True,
        xdim: str = "step",
        colorizer: Colorizer | None = default_colorizer,
        only2d: bool = False,
        fast=False,
        simple_offers_view=False,
    ):
        from negmas.plots.util import plot_mechanism_run

        return plot_mechanism_run(
            mechanism=self,
            negotiators=plotting_negotiators,
            save_fig=save_fig,
            path=path,
            fig_name=fig_name,
            ignore_none_offers=ignore_none_offers,
            with_lines=with_lines,
            show_agreement=show_agreement,
            show_pareto_distance=show_pareto_distance,
            show_nash_distance=show_nash_distance,
            show_kalai_distance=show_kalai_distance,
            show_max_welfare_distance=show_max_welfare_distance,
            show_max_relative_welfare_distance=show_max_relative_welfare_distance,
            show_end_reason=show_end_reason,
            show_annotations=show_annotations,
            show_reserved=show_reserved,
            colors=colors,
            markers=markers,
            colormap=colormap,
            ylimits=ylimits,
            common_legend=common_legend,
            extra_annotation="",
            xdim=xdim,
            colorizer=colorizer,
            show_total_time=show_total_time,
            show_relative_time=show_relative_time,
            show_n_steps=show_n_steps,
            only2d=only2d,
            fast=fast,
            simple_offers_view=simple_offers_view,
        )

    def add(
        self,
        negotiator: GBNegotiator,
        *,
        preferences: Preferences | None = None,
        role: str | None = None,
        ufun: BaseUtilityFunction | None = None,
    ) -> bool | None:
        added = super().add(negotiator, preferences=preferences, role=role, ufun=ufun)
        if added:
            for thread in self._threads:
                thread.responders.append(negotiator)

            thread = GBThread(
                self,  # type: ignore
                negotiator,
                responders=(
                    responders := [
                        _ for _ in self.negotiators if id(_) != id(negotiator)
                    ]
                ),
                indx=len(self._threads),
                evaluator=None,
                constraint=None,
                state=ThreadState(
                    new_offer=None,
                    new_responses={_.id: ResponseType.REJECT_OFFER for _ in responders},
                    accepted_offers=[],
                ),
            )
            self._threads.append(thread)
            self._current_state.threads[negotiator.id] = thread.state

        return added


class GBMechanism(BaseGBMechanism):
    def __init__(
        self,
        *args,
        evaluator_type: type[EvaluationStrategy] | None = None,
        evaluator_params: dict[str, Any] | None = None,
        local_evaluator_type: type[LocalEvaluationStrategy] | None = None,
        local_evaluator_params: dict[str, Any] | None = None,
        constraint_type: type[OfferingConstraint] | None = None,
        constraint_params: dict[str, Any] | None = None,
        local_constraint_type: type[LocalOfferingConstraint] | None = None,
        local_constraint_params: dict[str, Any] | None = None,
        response_combiner: Callable[[list[GBResponse]], GBResponse] = all_accept,
        dynamic_entry=False,
        extra_callbacks=False,
        check_offers=False,
        enforce_issue_types=False,
        cast_offers=False,
        end_on_no_response=True,
        ignore_negotiator_exceptions=False,
        parallel: bool = True,
        sync_calls: bool = False,
        initial_state: GBState | None = None,
        **kwargs,
    ):
        if not evaluator_type and not local_evaluator_type:
            raise ValueError(
                "You must pass either `evaluator_type` or `local_evaluator_type` to GBMechanism"
            )
        super().__init__(
            *args,
            dynamic_entry=dynamic_entry,
            extra_callbacks=extra_callbacks,
            initial_state=GBState() if not initial_state else initial_state,
            **kwargs,
        )
        self._current_state: GBState
        if not constraint_params:
            constraint_params = dict()
        self._global_constraint: OfferingConstraint | None = (
            instantiate(constraint_type, **constraint_params)
            if constraint_type
            else None
        )
        self._local_constraint_type: type[
            LocalOfferingConstraint
        ] | None = local_constraint_type
        self._local_constraint_params = (
            local_constraint_params if local_constraint_params else dict()
        )
        if not evaluator_params:
            evaluator_params = dict()
        self._global_evaluator = (
            instantiate(evaluator_type, **evaluator_params) if evaluator_type else None
        )
        self._local_evaluator_type = local_evaluator_type
        self._local_evaluator_params = (
            local_evaluator_params if local_evaluator_params else dict()
        )
        self._combiner = response_combiner
        self._parallel = parallel
        self._threads: list[GBThread] = []
        self._ignore_negotiator_exceptions = ignore_negotiator_exceptions
        self._dynamic_entry = dynamic_entry
        self._extra_callbacks = extra_callbacks
        self._check_offers = check_offers
        self._enforce_issue_types = enforce_issue_types
        self._cast_offers = cast_offers
        self._end_on_no_response = enforce_issue_types
        self._ignore_negotiator_exceptions = enforce_issue_types
        self._parallel = parallel
        self._sync_calls = sync_calls
        self.params["check_offers"] = check_offers
        self.params["enforce_issue_types"] = enforce_issue_types
        self.params["ignore_negotiator_exceptions"] = ignore_negotiator_exceptions
        self.params["cast_offers"] = cast_offers
        self.params["end_on_no_response"] = end_on_no_response
        self.params["enable_callbacks"] = extra_callbacks
        self.params["parallel"] = parallel
        self.params["sync_calls"] = sync_calls
        self.params["evaluator_type"] = get_full_type_name(evaluator_type)  # type: ignore
        self.params["evaluator_params"] = evaluator_params
        self.params["constraint_type"] = get_full_type_name(constraint_type)  # type: ignore
        self.params["constraint_params"] = constraint_params
        self.params["local_evaluator_type"] = get_full_type_name(local_evaluator_type)  # type: ignore
        self.params["local_evaluator_params"] = local_evaluator_params
        self.params["local_constraint_type"] = get_full_type_name(local_constraint_type)  # type: ignore
        self.params["local_constraint_params"] = local_constraint_params

    def add(
        self,
        negotiator: GBNegotiator,
        *,
        preferences: Preferences | None = None,
        role: str | None = None,
        ufun: BaseUtilityFunction | None = None,
    ) -> bool | None:
        added = super().add(negotiator, preferences=preferences, role=role, ufun=ufun)
        if added:
            for thread in self._threads:
                thread.responders.append(negotiator)
            evaluator, constraint = None, None
            if self._local_evaluator_type:
                evaluator = instantiate(
                    self._local_evaluator_type, **self._local_evaluator_params
                )
            if self._local_constraint_type:
                constraint = instantiate(
                    self._local_constraint_type, **self._local_constraint_params
                )

            thread = GBThread(
                self,
                negotiator,
                responders=(
                    responders := [
                        _ for _ in self.negotiators if id(_) != id(negotiator)
                    ]
                ),
                indx=len(self._threads),
                evaluator=evaluator,
                constraint=constraint,
                state=ThreadState(
                    new_offer=None,
                    new_responses={_.id: ResponseType.REJECT_OFFER for _ in responders},
                    accepted_offers=[],
                ),
            )
            self._threads.append(thread)
            self._current_state.threads[negotiator.id] = thread.state

        return added

    def set_sync_call(self, v: bool):
        self._sync_call = v

    def __call__(
        self, state: GBState, action: GBAction | None = None
    ) -> MechanismStepResult:
        # print(f"Round {self._current_state.step}")
        results = self.run_threads(action)
        # if state.step > self.outcome_space.cardinality:
        #     self.plot()
        #     breakpoint()
        if self.verbosity > 2:
            print(
                f"{self.name}: Central Processing after {humanize_time(perf_counter() - self._start_time, show_ms=True) if self._start_time else 0}",
                flush=True,
            )
        responses = []
        for source, (tstate, response) in results.items():
            state.threads[source] = tstate
            if self._local_evaluator_type:
                responses.append(response)
        if self.verbosity > 2:
            print(
                f"{self.name}: Global Evaluator after {humanize_time(perf_counter() - self._start_time, show_ms=True) if self._start_time else 0}",
                flush=True,
            )
        if self._global_evaluator:
            responses.append(
                self._global_evaluator(self.negotiator_ids, state, self._history)
            )
        if self.verbosity > 2:
            print(
                f"{self.name}: Global Constraint after {humanize_time(perf_counter() - self._start_time, show_ms=True) if self._start_time else 0}",
                flush=True,
            )
        if self._global_constraint:
            responses.append(self._global_constraint(state, self._history))

        e = self._combiner(responses)

        if e == "continue":
            return MechanismStepResult(state)
        if e is None:
            state.broken = True
        state.agreement = e
        return MechanismStepResult(state)

    @property
    def full_trace(self) -> list[TraceElement]:
        def response(state: GBState):
            if state.timedout:
                return "timedout"
            if state.ended:
                return "ended"
            if state.has_error:
                return "error"
            if state.agreement:
                return "accepted"
            return "rejected"

        offers = []
        self._history: list[GBState]  # type: ignore
        for state in self._history:
            state: GBState
            offers += [
                TraceElement(
                    state.time,
                    state.relative_time,
                    state.step,
                    source,
                    t.new_offer,
                    {neg: int(resp) for neg, resp in t.new_responses.items()},
                    response(state),
                )
                for source, t in state.threads.items()
            ]

        return offers

    @property
    def extended_trace(self) -> list[tuple[int, str, Outcome]]:
        """Returns the negotiation history as a list of step/negotiator/offer
        tuples."""
        offers = []
        self._history: list[GBState]  # type: ignore
        for state in self._history:
            offers += [
                (state.step, source, t.new_offer) for source, t in state.threads.items()
            ]
        return offers

    @property
    def trace(self) -> list[tuple[str, Outcome]]:
        """Returns the negotiation history as a list of negotiator/offer
        tuples."""
        offers = []
        for state in self._history:
            offers += [(source, t.new_offer) for source, t in state.threads.items()]

        return offers

    def negotiator_offers(self, negotiator_id: str) -> list[Outcome]:
        """Returns the offers given by a negotiator (in order)"""
        return [o for n, o in self.trace if n == negotiator_id]

    def negotiator_full_trace(
        self, negotiator_id: str
    ) -> list[tuple[float, float, int, Outcome, str]]:
        """Returns the (time/relative-time/step/outcome/response) given by a
        negotiator (in order)"""
        return [
            (t, rt, s, o, a)
            for t, rt, s, n, o, _, a in self.full_trace
            if n == negotiator_id
        ]

    @property
    def offers(self) -> list[Outcome]:
        """Returns the negotiation history as a list of offers."""
        return [o for _, o in self.trace]

    @property
    def _step(self):
        """A private property used by the checkpoint system."""
        return self._current_state.step

    def plot(
        self,
        plotting_negotiators: tuple[int, int] | tuple[str, str] = (0, 1),
        save_fig: bool = False,
        path: str | None = None,
        fig_name: str | None = None,
        ignore_none_offers: bool = True,
        with_lines: bool = True,
        show_agreement: bool = False,
        show_pareto_distance: bool = True,
        show_nash_distance: bool = True,
        show_kalai_distance: bool = True,
        show_max_welfare_distance: bool = True,
        show_max_relative_welfare_distance: bool = False,
        show_end_reason: bool = True,
        show_annotations: bool = False,
        show_reserved: bool = True,
        show_total_time=True,
        show_relative_time=True,
        show_n_steps=True,
        colors: list | None = None,
        markers: list[str] | None = None,
        colormap: str = DEFAULT_COLORMAP,
        ylimits: tuple[float, float] | None = None,
        common_legend: bool = True,
        xdim: str = "step",
        colorizer: Colorizer | None = default_colorizer,
        only2d: bool = False,
    ):
        from negmas.plots.util import plot_mechanism_run

        return plot_mechanism_run(
            mechanism=self,  # type: ignore
            negotiators=plotting_negotiators,
            save_fig=save_fig,
            path=path,
            fig_name=fig_name,
            ignore_none_offers=ignore_none_offers,
            with_lines=with_lines,
            show_agreement=show_agreement,
            show_pareto_distance=show_pareto_distance,
            show_nash_distance=show_nash_distance,
            show_kalai_distance=show_kalai_distance,
            show_max_welfare_distance=show_max_welfare_distance,
            show_max_relative_welfare_distance=show_max_relative_welfare_distance,
            show_end_reason=show_end_reason,
            show_annotations=show_annotations,
            show_reserved=show_reserved,
            colors=colors,
            markers=markers,
            colormap=colormap,
            ylimits=ylimits,
            common_legend=common_legend,
            extra_annotation="",
            xdim=xdim,
            colorizer=colorizer,
            show_total_time=show_total_time,
            show_relative_time=show_relative_time,
            show_n_steps=show_n_steps,
            only2d=only2d,
        )


class ParallelGBMechanism(GBMechanism):
    def __init__(self, *args, **kwargs):
        kwargs["parallel"] = True
        super().__init__(*args, **kwargs)


class SerialGBMechanism(GBMechanism):
    def __init__(self, *args, **kwargs):
        kwargs["parallel"] = False
        super().__init__(*args, **kwargs)



"""
Implements components used to consruct negotiators for SAO.
"""
from .base import *
from .acceptance import *
from .offering import *
from .concession import *
from .selectors import *
from .inverter import *
from .models import *

__all__ = (
    base.__all__
    + acceptance.__all__
    + offering.__all__
    + concession.__all__
    + selectors.__all__
    + inverter.__all__
    + models.__all__
)
